// MXResolver.cpp: implementation of the CMXResolver class.
//
//////////////////////////////////////////////////////////////////////
/*

  Usage:

	char *nameserver = "ifreeoffice.coasdfm";
	char *host = "hanmail.net";
	CMXResolver mxr;
	mxr.SetNameServer(nameserver);
	char *mailserver = mxr.MxQuery(host);
	if (mailserver == NULL) mailserver = host;
	printf("%s\r\n", mailserver);

*/


#include "MXResolver.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMXResolver::CMXResolver()
{

	m_best_mx[0] = '\0';
}

CMXResolver::~CMXResolver()
{

}

void CMXResolver::SetNameServer(char *server)
{
	m_server=server;
}
/*
RETURN	:
	The best MX host 
	NULL if not found any MX host from the DNS Server or something wrong with DNS Server.
*/
char *CMXResolver::MxQuery(char *domain)
{
	m_domain=domain;

	
	int sock;
	int numbytes, addr_len=0;
	
	struct hostent *he;
	struct sockaddr_in their_addr, from_addr;
	struct DNSHEADER dnsHeader;
	char buf[2048] = {0,};

	
	// by mark.  ʱȭ
	WSADATA _wsaData;
	if ( WSAStartup(MAKEWORD(2,1),&_wsaData) != 0 ){
		fprintf(stderr, "WSAStartup failed. code: %d\r\n", WSAGetLastError());
		return NULL;
	}
	/*
		Get IP Address of the nameserver.
	*/
	he = (struct hostent*) gethostbyname(m_server);
	if (he == NULL) {
		fprintf(stderr, "error gethostbyname. code=%d\r\n", WSAGetLastError());
		return NULL;
	}

	// create socket
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1) {
		// fprintf(stderr, "socket");
		return NULL;
	}
	
	/* address information to connect */
	their_addr.sin_family = AF_INET;
	their_addr.sin_port = htons(53);
	their_addr.sin_addr = *((struct in_addr *)he->h_addr);//INADDR_ANY;
	memset(their_addr.sin_zero,0,8);


	/* address information to receive */
	from_addr.sin_family = AF_INET;
	from_addr.sin_port = htons(53);
	from_addr.sin_addr.s_addr = INADDR_ANY;
	memset(from_addr.sin_zero,0,8);


	/* DNS Header information */
	dnsHeader.qryid		= htons(0x11DF);		// id
	dnsHeader.options	= 1;					// dns recursion
	dnsHeader.nqucount	= htons(1);				// 
	dnsHeader.nancount	= 0;
	dnsHeader.narcount	= 0;
	dnsHeader.naucount	= 0;
	dnsHeader.responses	= 0;


	int lenDH = sizeof(DNSHEADER);
	char dnsQuery[100]={0,};

	int posDnsQuery=0;

	memcpy(dnsQuery, &dnsHeader, lenDH);
	posDnsQuery += lenDH;

	// domain name ( as a QNAME )
	char qname[100]={0,};
	MakeQName(m_domain, qname);				// make a domain's query name

	memcpy(dnsQuery+posDnsQuery, qname, strlen(qname));
	posDnsQuery += strlen(qname);
	posDnsQuery ++;

	// mx type
	short iTemp;
	iTemp = htons(15);
	memcpy(dnsQuery+posDnsQuery, &iTemp, sizeof(short));
	posDnsQuery += sizeof(short);

	// inet class
	iTemp = htons(1);
	memcpy(dnsQuery+posDnsQuery, &iTemp, sizeof(short));
	posDnsQuery += sizeof(short);



	//@modify by mark at 2002/10/25
	// Ÿ ƿ 5ʷ Ѵ.

	int timeout_recv = 5000;
	int timeout_send = 5000;
	int bread;

	//  Ÿ ƿ
	bread = setsockopt( sock ,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout_recv, sizeof(timeout_recv));
	if(bread == SOCKET_ERROR) {
		fprintf(stderr,"failed to set recv timeout: %d\n",WSAGetLastError());
		return NULL;
	}

	// ۽ Ÿ ƿ
	bread = setsockopt( sock,SOL_SOCKET,SO_SNDTIMEO,(char*)&timeout_send, sizeof(timeout_send));
	if(bread == SOCKET_ERROR) {
		OutputDebugString("failed to set send timeout.\n");
		return NULL;
	}

	// send query
	if ((numbytes=sendto(sock, dnsQuery, posDnsQuery, 0,
		(struct sockaddr *)&their_addr, sizeof(struct sockaddr))) == -1) {
            // perror("sendto");
            // exit(1);
			return NULL;
        }
	// ۽ Ÿ ƿ üũ
	int bwrote = numbytes;
	if (bwrote == SOCKET_ERROR){
		if (WSAGetLastError() == WSAETIMEDOUT) {
			OutputDebugString("mx query timed out for sending data to ... \n");
			return NULL;
		}
		OutputDebugString("sendto failed\n");
		return NULL;
	}

	
	// receive query
	addr_len = sizeof(struct sockaddr);
#ifdef _WIN32
	numbytes = recvfrom(sock, buf, 2048, 0, (struct sockaddr *)&from_addr, &addr_len);
#else
	numbytes = recvfrom(sock, buf, 2048, 0, (struct sockaddr *)&from_addr, (socklen_t*)&addr_len);
#endif
	
	//  Ÿ ƿ üũ
	bread = numbytes;
	if (bread == SOCKET_ERROR) {
		if (WSAGetLastError() == WSAETIMEDOUT) {
			OutputDebugString("mx query timed out for recv from.\n");
			return NULL;
		}
		OutputDebugString("recvfrom failed.\n");
		return NULL;
	}

	//@check Ʒ ڵ  ʿ  ʳ?
	if (numbytes == -1) {
		//perror("recvfrm");
		//PrintLastError();
		//exit(1);
		return NULL;
	}

//	printf("got packet from %s\n",inet_ntoa(their_addr.sin_addr));
//	printf("packet is %d bytes long\n",numbytes);
	
	GetMXName(buf, m_best_mx);
	//printf("\r\nBest MX = %s\r\n", m_best_mx);

	if (*m_best_mx == '\0') return NULL;
	else return m_best_mx;
}







/*

  ParseName()

DESC	:
  Gets the domain name in the responsed packet from DNS server.
  ( Decompresses the domain name if it as compressed. )
INPUT	:
  org_buf	- the packet buffer,
  idx		- position (offset) of domain name to parse from the beggining of the packet.
  sName		- a pointer to receive the domain name.
RETURN	:
  a pointer of the end of the domain name (which will be parsed by this function)
	in packet buffer.
*/
char *CMXResolver::ParseName (char *org_buf, int idx, char *sName) {

	bool bCompressed = false;
	unsigned char ch;
//	int iLenName=0;

	char *ret=NULL;
	char *ptr=sName;
	char *buf = org_buf + idx;

	while (*buf) {
		ch = *buf ++;
		if ( ch >= 192 ) {	// is compressed domain ?
			bCompressed = true;		// then, check it.

			{	// and check the return pointer.
				// WARNING: what if the domain is compress as recursively? 
				// Well, I'm not sure what will happen.
				// You should do something here in that case.
				ret = buf + 1;
			}
			

			/*
				If compressed, the pointer (offset) of original domain lable part
				will be held in the last 6 bit of the first octet of the length and next octet.
			*/
			ch = ch - 192;
			buf = org_buf + ch + *buf;
			// buf = org_buf + *buf;		//<!-- DPRECATED -->
			ch = *buf ++;		// skip length field.
		}
		for (int k=0; k<ch; k++) {
			*ptr++ = *buf ++;
		}
		if (*buf) *ptr ++ = '.';
	}
	*ptr='\0';

	if (bCompressed) return ret; //org_buf + idx + 2; // if compressed
	else return ++buf;
}

/*

  GetMXName()
DESC	:
  Parses the response (packet) of DNS Server.
  Make sure that this function (parsing) is used for a MX Query.
INPUT	:
	buf		- packet buffer received from DNS server
	ret		- a pointer to receive the *best mx host name*
RETURN	:
	the best mx host name as in 'ret' variable (call by referrence)
	If there is no mx query result in packet, then it returns nothing.
*/
void CMXResolver::GetMXName(char *buf, char *ret) {

//	int iChCount=0;		// character counter
	char qName[100]={0,};

	int lowest=-1;
	char *nbuf=buf;

	short qType, qClass, qRDLen=0, qPreferrence=0;
	long qTTL;


//	printf("\r\n[ Packet header ]\r\n");
	nbuf += 6;								// pos of num of the results in header
	
	short nResult=0;
	memcpy(&nResult, nbuf, sizeof(short));
	nResult = ntohs(nResult);				// num of totoal retults
	if (nResult == 0) return;

//	printf("total results=%d\r\n", nResult);

	nbuf += 6;								// pos of end of header

//	printf("\r\n[ Question section ]\r\n");
	nbuf = ParseName (buf, (nbuf-buf), qName); // get domain name in question section
//	printf("domain = %s\r\n", qName);

	// query type
	memcpy(&qType, nbuf, sizeof(short));
	qType = ntohs(qType);
//	printf("query type = %d\r\n", qType);
	nbuf += 2;			// skip query type

	// query class
	memcpy(&qClass, nbuf, sizeof(short));
	qClass = ntohs(qClass);
//	printf("query class = %d\r\n", qClass);
	nbuf += 2;			// skip query class


	

	
	// now from here, answer section
//	printf("\r\n[ Answer section ]\r\n");
	while (nResult -- > 0) {
		nbuf = ParseName(buf, (nbuf-buf), qName);
		//printf("domain = %s\r\n", qName);

		// query type
		memcpy(&qType, nbuf, sizeof(short));
		qType = ntohs(qType);
		//printf("query type = %d\r\n", qType);
		nbuf += 2;			// skip query type

		// query class
		memcpy(&qClass, nbuf, sizeof(short));
		qClass = ntohs(qClass);
		//printf("query class = %d\r\n", qClass);
		nbuf += 2;			// skip query class

		// TTL
		memcpy(&qTTL, nbuf, sizeof(long));
		qTTL = ntohl(qTTL);
		//printf("TTL = %d\r\n", qTTL);
		nbuf += 4;			// skip TTL

		if (qType == 15) {	// MX Query?
			// data length
			memcpy(&qRDLen, nbuf, sizeof(short));
			qRDLen = ntohs(qRDLen);
			//printf("RData Length = %d\r\n", qRDLen);
			nbuf += 2;			// skip
			// get preferrence
			memcpy(&qPreferrence, nbuf, sizeof(short));
			qPreferrence = ntohs(qPreferrence);
//			printf("Preferrence [%d] ", qPreferrence);
			nbuf += 2;			// skip

			// get data
			nbuf = ParseName(buf, (nbuf-buf), qName);
//			printf("mx=%s\r\n", qName);


			// check & get the best mx
			if (lowest == -1 || lowest > qPreferrence) {
				lowest = qPreferrence;
				strcpy(ret, qName);
			}
		}
	}
}


void CMXResolver::PrintLastError()
{

#ifdef _WIN32
	int e = WSAGetLastError();
	switch (e) {
		case WSANOTINITIALISED :
			printf("WSANOTINITIALISED");
			break;
		case WSAENETDOWN :
			printf("WSAENETDOWN");
			break;
		case WSAEFAULT :
			printf("WSAEFAULT\r\n");
			break;
		case WSAEINTR :
			printf("WSAEINTR");
			break;
		case WSAEINPROGRESS :
			printf("WSAEINPROGRESS");
			break;
		case WSAEINVAL :
			printf("WSAEINVAL");
			break;
		case WSAEISCONN :
			printf("WSAEISCONN");
			break;
		case WSAENETRESET :
			printf("WSAENETRESET");
			break;
		case WSAENOTSOCK :
			printf("WSAENOTSOCK");
			break;
		case WSAEOPNOTSUPP :
			printf("WSAEOPNOTSUPP");
			break;
		case WSAESHUTDOWN :
			printf("WSAESHUTDOWN");
			break;
		case WSAEWOULDBLOCK :
			printf("WSAEWOULDBLOCK");
			break;
		case WSAEMSGSIZE :
			printf("WSAEMSGSIZE");
			break;
		case WSAETIMEDOUT :
			printf("WSAETIMEDOUT");
			break;
		case WSAECONNRESET :
			printf("WSAECONNRESET");
			break;
	}
#endif
}

/*

MakeQName()
INPUT	:
	domain	- domain name to make as a query name for DNS request.
	qname	- pointer to receive the query name

NOTE	:

Takes sDomain and converts it to the QNAME-type string, returns that. QNAME is how a
DNS server expects the string.

 Ex...    Pass -        mail.com
          Returns -     &H4mail&H3com
                        ^      ^
                        |______|____ These two are character counters, they count the
                                     number of characters appearing after them
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
*/
void CMXResolver::MakeQName(const char *domain, char *qname)
{	
	char ch;
	char sDotName[254]={0,};
	int		iLen = strlen(domain);
	int		iLenDot=0;
	char	*dst	= qname;


	for (int i=0; i < iLen; i++) {
		ch = *(domain + i);
		if (ch == '.') {
			*dst++ = (char)iLenDot;
			for(int j=0; j < iLenDot; j++) {
				*dst++=sDotName[j];
			}
			iLenDot=0;
		}
		else {
			sDotName[iLenDot++] = ch;
		}
	}
	*dst ++ = (char) iLenDot;
	for(int j=0; j < iLenDot; j++) {
		*dst++=sDotName[j];
	}
}
